###############################################################################
#
# R source file to accompany "Modelling mortality and longevity with actuarial
# applications", by Macdonald, A. S., Richards, S. J. and Currie, I. D.,
# Cambridge University Press.
#
# This file provides commonly used functions for fitting survival models.

###############################################################################
#
# Function to fit survival model with left-truncated and right-censored data.
# Requires appropriate definition of NegLogL() in same R workspace.
fitSurvivalModel = function(vData, parnames, parvalues)
{
  # Call nlm() function to minimize the negative log-likelihood, -l.  The user
  # needs to supply suitable starting values in parvalues.  We need the Hessian
  # matrix for calculating standard errors, so we must specify hessian=T.
  # Finally, we must give the optimiser a clue of the relative order of each
  # parameter, so we set typsize to parvalues as well.
  Estimates = nlm(NegLogL, p=parvalues, vData, typsize=parvalues, hessian=T, stepmax=1, iterlim=200)

  npars = length(parvalues)
  dLogLikelihood = -Estimates$minimum
  dAIC = -2*dLogLikelihood + 2*length(parvalues)

  # The variance-covariance matrix of the parameter estimates is the inverse of
  # the Hessian of -l.
  VarCov = solve(Estimates$hessian, diag(npars))

  sNLM = sprintf("nlm() return code %d.\n", Estimates$code)
  if (1 != Estimates$code)
    sNLM = paste0(sNLM, "Warning! nlm() did not return cleanly.\n")

  # Now paste together the outputs into a string, which can be either echoed to
  # the console or written to a file.
  format = paste0("Data cover %d lives and %d deaths.\n",
                  "Total exposure time is %.2f years.\n\n",
                  "Iterations=%d\n",
                  "Log-likelihood=%.2f\n",
                  "AIC=%.2f\n\n",
                  "Parameter    Estimate Std. error\n",
                  "------------ -------- ----------\n")
  sSummary = sprintf(format,
                     length(vData$d), sum(vData$d),
                     sum(vData$TimeObserved),
                     Estimates$iterations, dLogLikelihood,
                     dAIC)
  for (i in 1:npars)
    sSummary = paste0(sSummary, display(Estimates, VarCov, parnames, i))

  return(list(summary=paste0(sNLM, sSummary), pars=Estimates$estimate))
}

###############################################################################
#
# Helper function to extract parameter estimate and standard error and prepare
# a suitable string for printing.
display = function(Est, VarCov, parnames, i)
{
  # Pad out the names in case less than nine characters
  parnames = paste0(parnames, "         ")
  parnames = substr(parnames, 1, 12)
  return(paste0(parnames[i], " ", formatC(Est$estimate[i], format="fg", digits=6, width=8),
         " ", formatC(sqrt(VarCov[i,i]), format="fg", digits=6, width=10), "\n"))
}

###############################################################################
#
# Function to split up the individual time lived into single years of age.
# Note that age zero is excluded, i.e. experience is only split up over the age
# range from 1 to uiMaxAge.  Returns deaths (dx), central exposures (Ecx) and
# the amounts-weighted equivalents (dxa and Ecxa)
splitExperienceByAge = function(vData, uiMaxAge)
{
  ages = 1:uiMaxAge
  dx   = rep(0, uiMaxAge)
  dxa  = rep(0.0, uiMaxAge)
  Ecx  = rep(0.0, uiMaxAge)
  Ecxa = rep(0.0, uiMaxAge)

  vData$ExitAge = vData$EntryAge+vData$TimeObserved
  for (i in (1:length(vData$EntryAge)))
  {
    x=floor(vData$ExitAge[i])
    dx [x] = dx [x]+vData$d[i]
    dxa[x] = dxa[x]+vData$d[i]*vData$Benefit[i]
    for (x in (1:uiMaxAge))
    {
      dExposure = min(x+1, vData$ExitAge[i]) - max(x, vData$EntryAge[i])
      if (dExposure > 0.0)
      {
        Ecx [x] = Ecx [x]+dExposure
        Ecxa[x] = Ecxa[x]+dExposure*vData$Benefit[i]
      }
    }
  }
  return(list(x=ages, dx=dx, dxa=dxa, Ecx=Ecx, Ecxa=Ecxa))
}

###############################################################################
#
# Function to calculate bootstrap test statistics for integrated hazard
# function H(), which is presumed defined in the same R workspace and has
# access to the current values of the parameters.  See Section 7.7 of the book.
bootstrap = function(vData, bcount, samplesize)
{
  vdAElives = c()
  vdAEamounts = c()
  for (i in 1:bcount)
  {
    # Generate indices of cases to look up
    cases = sample(1:length(vData$d), samplesize, replace=TRUE)
    tData = vData[cases,]
    Hxt = H(tData)
    dAElives = sum(tData$d) / sum(Hxt)
    dAEamounts = sum(tData$Benefit*tData$d) / sum(tData$Benefit*Hxt)
    vdAElives = c(vdAElives, dAElives)
    vdAEamounts = c(vdAEamounts, dAEamounts)
  }
  return(list(AElives=vdAElives, AEamounts=vdAEamounts))
}

###############################################################################
#
# Function to calculate the integrated hazard over single years of age.
# Note that age zero is excluded, i.e. experience is only split up over the age
# range from 1 to uiMaxAge.  Returns deaths (dx) and integrated hazards (Hx).
calculateDevianceResiduals = function(vData, uiMaxAge, oModel)
{
  ages = 1:uiMaxAge
  dx   = rep(0, uiMaxAge)
  Hx   = rep(0.0, uiMaxAge)

  vData$ExitAge = vData$EntryAge+vData$TimeObserved
  for (i in (1:length(vData$EntryAge)))
  {
    dAlpha = vData$Alpha[i]
    dBeta  = vData$Beta[i]
    x=floor(vData$ExitAge[i])
    dx[x] = dx[x]+vData$d[i]
    for (x in (1:uiMaxAge))
    {
      dEntry = max(x, vData$EntryAge[i])
      dExit = min(x+1, vData$ExitAge[i])
      t = dExit-dEntry
      if (t > 0.0)
        Hx[x] = Hx[x] + exp(dAlpha + dBeta*dEntry) * (exp(dBeta * t) - 1.0) / dBeta
    }
  }
  DevRes <- rep(0.0, uiMaxAge)
  for (i in 1:uiMaxAge)
  {
    O = dx[i]
    E = Hx[i]
    temp = 0.0
    if (O>0)
      temp = O*log(O/E)

    DevRes[i] = sign(O-E)*sqrt(2*(temp-(O-E)))
  }
  return(list(x=ages, dx=dx, Hx=Hx, DevRes=DevRes))
}
