#
#    Forecasting with the Lee-Carter model
#
#    rm(list = ls())
#
#    Get some data 
#
source("Read_HMD.r")
#
Death.Data = Read.HMD("Data/HMD_UK_Deaths_2013.txt")
names(Death.Data)
Exposure.Data = Read.HMD("Data/HMD_UK_Exposures_2013.txt")
names(Exposure.Data)
#
Age = Death.Data$Age; Age
Year = Death.Data$Year; Year
#
#    Choose male data for ages 50 to 90 and years 1961 to 2013
#
AGE = 50:90
YEAR = 1961:2013 
Dth = Death.Data$Male.Matrix[(Age > 49) & (Age < 91), Year > 1960]
Exp = Exposure.Data$Male.Matrix[(Age > 49) & (Age < 91), Year > 1960]
Obs = log(Dth/Exp)
n.x = nrow(Dth)
n.y = ncol(Dth)
#
#     Fit model with gnm function
#
library("gnm", lib.loc = "/home/iain/reml/R.LIBRARIES")
Age.F = factor(rep(AGE, n.y))          #  Make age a factor
Year.F = factor(rep(YEAR, each = n.x)) #  Make year a factor
Dth.V = c(Dth)   #  Make Dth a vector
Exp.V = c(Exp)   #  Make exp a vector
LC.Male = gnm(Dth.V ~ -1 + Age.F + Mult(Age.F, Year.F),
                offset = log(Exp.V), family = poisson)
#
#     Get coefficients
#
Alpha.gnm = LC.Male$coef[1:n.x]
Beta.gnm = LC.Male$coef[(n.x+1):(2*n.x)]
Kappa.gnm = LC.Male$coef[(2*n.x+1):(2*n.x+n.y)]
#
#    Convert to standard Lee-Carter parameters and check 
#      1) location and scale constraints
#      2) invariance of fitted log mortality
#
Kappa.m = mean(Kappa.gnm)
Beta.m = mean(Beta.gnm)
Alpha.hat = Alpha.gnm + Kappa.m*Beta.gnm
Beta.hat = Beta.gnm/(n.x * Beta.m)
Kappa.hat = n.x * Beta.m*(Kappa.gnm - Kappa.m)
Fitted.M.hat = Alpha.hat + Beta.hat %*% t(Kappa.hat)
#
c(sum(Kappa.hat), sum(Beta.hat))                     # 1) OK
max(abs(c(Fitted.M.hat) - log(LC.Male$fit/Exp.V)))   # 2) OK
#
#      Forecast kappa with drift model
#
library("astsa", lib.loc = "/home/iain/reml/R.LIBRARIES")
N.Ahead = 25
Kappa.for = sarima.for(Kappa.hat, n.ahead = N.Ahead, p=0, d=1, q=0)
names(Kappa.for)
# 
#      Prediction error
#
Central = Kappa.for$pred
SE.Pred = Kappa.for$se
Z = qnorm(0.975)
Kappa.Up.Pred = Central + Z*SE.Pred
Kappa.Dn.Pred = Central - Z*SE.Pred
Forecast = Alpha.hat + Beta.hat %*% t(Central)
Forecast.Up.Pred = Alpha.hat + Beta.hat %*% t(Kappa.Up.Pred)
Forecast.Dn.Pred = Alpha.hat + Beta.hat %*% t(Kappa.Dn.Pred)
Range = 2014:2038
#
#   Graphical output
#
Plot.Age = 65
Plot.Row = Plot.Age - min(AGE) + 1
par(mfrow = c(1,1))
par(mar=c(4.2, 4, 1, 0.5), mgp=c(3, 1, 0), las=1, cex=1)
Year.For = seq(YEAR[1], (YEAR[length(YEAR)]+25))
plot(YEAR + 0.5, Obs[Plot.Row, ], axes = FALSE, xlab = "Year", ylab = "log(mortality)",
     xlim = c(1961, 2040),
     ylim = c(min(Forecast.Dn.Pred[Plot.Row, ]), max(Obs[Plot.Row, ]) ))
axis(1,las = 1, at = seq(1950, 2040, by = 10), tcl = -0.4)
axis(2, seq(-6,-2, by = 0.5), tcl = -0.4)
lines(Range + 0.5, Forecast[Plot.Row, ], lwd = 2)
lines(Range + 0.5, Forecast.Up.Pred[Plot.Row, ], lty = 2, lwd = 2)
lines(Range + 0.5, Forecast.Dn.Pred[Plot.Row, ], lty = 2, lwd = 2)
legend("bottomleft", legend = c("Observed", "Central forecast", "95% PI"),
       lty = c(-1, 1, 2, 2), pch = c(1,-1,-1,-1), lwd = 2, bty = "n") 
# 
#      Parameter error
#
Sarima.out = sarima(Kappa.hat, p=0, d=1, q=0, details = FALSE)
SE.Param = (1:N.Ahead) * sqrt(Sarima.out$fit$sigma2/(n.y-1))
Kappa.Up.Param = Central + Z*SE.Param
Kappa.Dn.Param = Central - Z*SE.Param
Forecast = Alpha.hat + Beta.hat %*% t(Central)
Forecast.Up.Param = Alpha.hat + Beta.hat %*% t(Kappa.Up.Param)
Forecast.Dn.Param = Alpha.hat + Beta.hat %*% t(Kappa.Dn.Param)
#
par(mar=c(4.2, 4, 1, 0.5), mgp=c(3, 1, 0), las=1, cex=1)
plot(YEAR+0.5, Obs[Plot.Row, ], axes = FALSE, xlab = "Year", ylab = "log(mortality)",
     xlim = c(1961, 2040),
     ylim = c(min(Forecast.Dn.Param[Plot.Row, ]), max(Obs[Plot.Row, ]) ))
axis(1,las = 1, at = seq(1950, 2040, by = 10), tcl = -0.4)
axis(2, seq(-6,-3, by = 0.5), tcl = -0.4)
lines(Range+0.5, Forecast[Plot.Row, ], lwd = 2)
lines(Range+0.5, Forecast.Up.Param[Plot.Row, ], lty = 2, lwd = 2)
lines(Range+0.5, Forecast.Dn.Param[Plot.Row, ], lty = 2, lwd = 2)
legend("bottomleft", legend = c("Observed", "Central forecast", "95% CI"),
       lty = c(-1, 1, 2, 2), pch = c(1,-1,-1,-1), lwd = 2, bty = "n") 
