#
#    Fits the classical Gompertz model
#
#    Load the HMD read function Read_HMD from the file Read_HMD.r
#
#    rm(list = ls())   #   Optionally empty workspace
#
source("Read_HMD.r")
#
#    Get some data from the directory Data
#
Death.Data = Read.HMD("Data/HMD_UK_Deaths.txt")
names(Death.Data)
Exposure.Data = Read.HMD("Data/HMD_UK_Exposures.txt")
names(Exposure.Data)
#
Age = Death.Data$Age; Age
Year = Death.Data$Year; Year
#
#    Choose male data for ages 40-90 for year 2000
#
AGE = 40:90
Dth = Death.Data$Male.Matrix[(39 < Age) & (Age < 91), Year == 2000]
Exp = Exposure.Data$Male.Matrix[(39 < Age) & (Age < 91), Year == 2000]
cbind(AGE, Dth, Exp)
#
#    Unweighted least squares with log \mu
#
Fit.Gompertz = lm(log(Dth/Exp) ~ AGE)
summary(Fit.Gompertz)
#
par(mar=c(4.2, 4, 0, 0.5), mgp=c(3, 1, 0), las=1, cex=1)
AGE.Plot = AGE + 1/2
plot(x = AGE.Plot, y = log(Dth/Exp), type = "n", xlab = "Age", 
     ylab = expression(paste("log(", mu,")")), axes = FALSE)
axis(1, las = 1, at = seq(30,90,10)); axis(2, las = 1, at = seq(-7,-1,1))
points(x = AGE.Plot, y = log(Dth/Exp))
lines(AGE.Plot, Fit.Gompertz$fit)
legend("bottomright", pch=c(1, -1), lty = c(-1,1),
       legend=c("Observed", "Fitted"), bty="n")
#
#   Least squares fomulae
#
n = length(AGE)
One = rep(1, n)
X = cbind(One, AGE)
y = log(Dth/Exp)
theta = solve(t(X) %*% X, t(X) %*% y)
cbind(theta, Fit.Gompertz$coef)
y.hat = X %*% theta
Sigma2 = c(t(y - y.hat) %*% (y - y.hat)) / (n - 2)
#
#   Confidence intervals
#
Var.y.hat = Sigma2 * X %*% solve(t(X) %*% X, t(X)) 
SD.y.hat = c(sqrt(diag(Var.y.hat)))
CI.up = y.hat + qt(0.975, n - 2) * SD.y.hat
CI.dn = y.hat + qt(0.025, n - 2) * SD.y.hat
Fit.and.CI = predict(Fit.Gompertz, interval="confidence")
cbind(Fit.and.CI, y.hat, CI.dn, CI.up)[1:3, ]
#
#   Prediction intervals
#
Var.y.hat.pred = Sigma2 * (diag(n) + X %*% solve(t(X) %*% X, t(X)))
SD.y.hat.pred = c(sqrt(diag(Var.y.hat.pred)))
PI.up = y.hat + qt(0.975, n - 2) * SD.y.hat.pred
PI.dn = y.hat + qt(0.025, n - 2) * SD.y.hat.pred
Fit.and.PI = predict(Fit.Gompertz, interval="prediction")
cbind(Fit.and.PI, y.hat, PI.dn, PI.up)[1:3, ]
#
#    Gompertz model as a GLM with Poisson errors
#
Gompertz.P = glm(Dth ~ AGE + offset(log(Exp)), family = poisson)
summary(Gompertz.P)
#
#   Compare fitted values & standard errors for normal and Poisson models
#
Normal.Reg = predict(Fit.Gompertz, se.fit = TRUE)
Poisson.GLM = predict(Gompertz.P, se.fit = TRUE)
cbind(Normal.Reg$fit, Poisson.GLM$fit - log(Exp))[1:3, ]
cbind(Normal.Reg$se.fit, Poisson.GLM$se.fit)[1:3, ]
#

