#
#    Simulation
#
#    This program illustrates simulation with the Lee-Carter model
#    We simulate kappa in the L-C model with the drift model.
#    Parameter error, stochastic error and joint parameter and stochastic 
#    error are illustrated.
#
#    rm(list = ls())
#
#    Get some data and fit the L-C model as before
#
source("Read_HMD.r")
#
Death.Data = Read.HMD("Data/HMD_UK_Deaths_2013.txt")
names(Death.Data)
Exposure.Data = Read.HMD("Data/HMD_UK_Exposures_2013.txt")
names(Exposure.Data)
#
Age = Death.Data$Age; Age
Year = Death.Data$Year; Year
#
#    Choose male data for ages 50 to 90 and years 1961 to 2013
#
AGE = 50:90
YEAR = 1961:2013 
Dth = Death.Data$Male.Matrix[(Age > 49) & (Age < 91), Year > 1960]
Exp = Exposure.Data$Male.Matrix[(Age > 49) & (Age < 91), Year > 1960]
Obs = log(Dth/Exp)
n.x = nrow(Dth)
n.y = ncol(Dth)
#
#     Fit model with gnm function
#
library("gnm", lib.loc = "/home/iain/reml/R.LIBRARIES")
Age.F = factor(rep(AGE, n.y))          #  Make age a factor
Year.F = factor(rep(YEAR, each = n.x)) #  Make year a factor
Dth.V = c(Dth)   #  Make Dth a vector
Exp.V = c(Exp)   #  Make exp a vector
LC.Male = gnm(Dth.V ~ -1 + Age.F + Mult(Age.F, Year.F),
                offset = log(Exp.V), family = poisson)
#
#     Get coefficients
#
Alpha.gnm = LC.Male$coef[1:n.x]
Beta.gnm = LC.Male$coef[(n.x+1):(2*n.x)]
Kappa.gnm = LC.Male$coef[(2*n.x+1):(2*n.x+n.y)]
#
#    Convert to standard Lee-Carter parameters and check 
#      1) location and scale constraints
#      2) invariance of fitted log mortality
#
Kappa.m = mean(Kappa.gnm)
Beta.m = mean(Beta.gnm)
Alpha.hat = Alpha.gnm + Kappa.m*Beta.gnm
Beta.hat = Beta.gnm/(n.x * Beta.m)
Kappa.hat = n.x * Beta.m*(Kappa.gnm - Kappa.m)
Fitted.M.hat = Alpha.hat + Beta.hat %*% t(Kappa.hat)
#
c(sum(Kappa.hat), sum(Beta.hat))                     # 1) OK
max(abs(c(Fitted.M.hat) - log(LC.Male$fit/Exp.V)))   # 2) OK
#
#     With astsa
#
# install.packages("astsa", lib = "/home/iain/reml/R.LIBRARIES")
library("astsa", lib.loc = "/home/iain/reml/R.LIBRARIES")
#
#     Drift estimates - use package astsa and direct calculation
#
Kappa = Kappa.hat
Year = YEAR
n.y = length(Kappa)
Sarima.out = sarima(Kappa, p=0, d=1, q=0, details = FALSE)
Mu = mean(diff(Kappa))
c(Mu, Sarima.out$fit$coef)                         #   Check
Sigma2 = var(diff(Kappa))
c((n.y-2)/(n.y-1)*Sigma2, Sarima.out$fit$sigma2)   #   Check
#
#     Forecast with the sarima.for function
#
N.Ahead = 25
Forecast = sarima.for(Kappa, n.ahead = N.Ahead, p=0, d=1, q=0)
#
#     ==========================================================
#
#     Simulation
#
#     +++++++++++    Stochastic error     ++++++++++++++
#
Sigma = sqrt(Sigma2)
set.seed(1)
Range = max(Year+1):max(Year + N.Ahead)
N.Sim = 100
#   
#    The idea is to create a matrix with N.Sim rows, each row is a simulation.
#    Then the apply function with the cumsum function is used to create the
#    N.Sim forecasts.  These can be plotted with the matplot function.
#
Mat.Sim = matrix(rnorm(N.Sim * N.Ahead, Mu, Sigma), nrow = N.Sim, ncol = N.Ahead)
Stoc = Kappa[n.y] + apply(Mat.Sim, 1, cumsum)
par(mar=c(4.2, 4, 1, 0.5), mgp=c(3, 1, 0), las=1, cex=1)
matplot(Range, Stoc, type = "l", col = 1, xlab = "Year", ylab = expression(kappa),
        axes = FALSE)
axis(1, las = 1, at = seq(2010, 2040, by = 5), tcl = -0.4)
axis(2, seq(-70, -10, by = 5), tcl = -0.4)
#
#    Now let's compare the simulated and theoretical results
#
set.seed(1)
N.Sim = 1000
Mat.Sim = matrix(rnorm(N.Sim * N.Ahead, Mu, Sigma), nrow = N.Sim, ncol = N.Ahead)
Stoc = Kappa[n.y] + apply(Mat.Sim, 1, cumsum)
#
#     Define a function to calculate the 2.5%, 50% & 97.5% quantiles.
#     This enables us the compare the theoretical PIs with the simulated
#     First the simulated quantiles
#
Quantile = function(x) quantile(x, c(0.025, 0.5, 0.975))
Sim.Q = apply(Stoc, 1, Quantile)
par(mar=c(4.2, 4, 1, 0.5), mgp=c(3, 1, 0), las=1, cex=1)
plot(Range, Sim.Q[2, ], type = "l", lwd = 2, axes = FALSE,
     xlab = "Year", ylab = expression(kappa), ylim = range(Sim.Q, Sim.Q))
axis(1, las = 1, at = seq(2010, 2040, by = 5), tcl = -0.4)
axis(2, seq(-70, -10, by = 5), tcl = -0.4)
lines(Range, Sim.Q[1, ], lwd = 2)
lines(Range, Sim.Q[3, ], lwd = 2)
#
#     And now the theoretical
#
Central = Forecast$pred
SE.Stoch = Forecast$se
PI.Up = Central + 1.96*SE.Stoch
PI.Dn = Central - 1.96*SE.Stoch
lines(Range, Central, lwd = 2, lty = 2)
lines(Range, PI.Up, lwd = 2, lty = 2)
lines(Range, PI.Dn, lwd = 2, lty = 2)
legend("bottomleft", legend = c("95% PI: theoretical", "95% PI: simulated"),
       lty = c(2,1), lwd = 2, bty = "n")
#
#     +++++++++++++      Parameter error     +++++++++++++++
#
#     This uses the same ideas as the previous section but simulates
#     from the distribution of the drift parameter, mu
#
set.seed(1)
N.Sim = 50
Mat.Sim = matrix(rep(rnorm(N.Sim, Mu, Sigma/sqrt(n.y-1)), N.Ahead),
  nrow = N.Sim, ncol = N.Ahead)
Param = Kappa[n.y] + apply(Mat.Sim, 1, cumsum)
par(mar=c(4.2, 4, 1, 0.5), mgp=c(3, 1, 0), las=1, cex=1)
matplot(Range, Param, type = "l", col = 1, xlab = "Year",
        ylab = expression(kappa), axes = FALSE)
axis(1, las = 1, at = seq(2010, 2040, by = 5), tcl = -0.4)
axis(2, seq(-70, -10, by = 5), tcl = -0.4)
#
set.seed(1)
N.Sim = 5000
Mat.Sim = matrix(rep(rnorm(N.Sim, Mu, Sigma/sqrt(n.y-1)), N.Ahead),
  nrow = N.Sim, ncol = N.Ahead)
Param = Kappa[n.y] + apply(Mat.Sim, 1, cumsum)
Sim.Q.Param = apply(Param, 1, Quantile)
par(mar=c(4.2, 4, 1, 0.5), mgp=c(3, 1, 0), las=1, cex=1)
plot(Range, Sim.Q.Param[2, ], type = "l", lwd = 2, axes = FALSE,
     xlab = "Year", ylab = expression(kappa),
     ylim = range(Sim.Q.Param, Sim.Q.Param))
axis(1, las = 1, at = seq(2010, 2040, by = 5), tcl = -0.4)
axis(2, seq(-70, -10, by = 5), tcl = -0.4)
lines(Range, Sim.Q.Param[1, ], lwd = 2)
lines(Range, Sim.Q.Param[3, ], lwd = 2)
#
lines(Range, Central, lwd = 2, lty = 2)
SE = (1:N.Ahead) * sqrt(Sarima.out$fit$sigma2/(n.y-1))
CI.Up = Central + 1.96*SE
CI.Dn = Central - 1.96*SE
lines(Range, CI.Up, lty = 2, lwd = 2)
lines(Range, CI.Dn, lty = 2, lwd = 2)
legend("bottomleft", legend = c("95% CI: theoretical", "95% CI: simulated"),
       lty = c(2,1), lwd = 2, bty = "n")
#
#    +++++++++++   Stochastic and parameter error     +++++++++++
#
set.seed(1)
N.Sim = 100
#   
#    We simulate the stochastic error with mean zero
#    and then add the simulated values of mu
#
Mat.Sim.1 = matrix(rnorm(N.Sim * N.Ahead, 0, Sigma),
  nrow = N.Sim, ncol = N.Ahead)
Mat.Sim.2 = matrix(rep(rnorm(N.Sim, Mu, Sigma/sqrt(n.y-1)), N.Ahead),
  nrow = N.Sim, ncol = N.Ahead)
Mat.Sim = Mat.Sim.1 + Mat.Sim.2
Both = Kappa[n.y] + apply(Mat.Sim, 1, cumsum)
# 
par(mar=c(4.2, 4, 1, 0.5), mgp=c(3, 1, 0), las=1, cex=1)
matplot(Range, Both, type = "l", col = 1, xlab = "Year",
        ylab = expression(kappa), axes = FALSE)
axis(1, las = 1, at = seq(2010, 2040, by = 5), tcl = -0.4)
axis(2, seq(-70, -10, by = 5), tcl = -0.4)
#
set.seed(1)
N.Sim = 1000
Mat.Sim.1 = matrix(rnorm(N.Sim * N.Ahead, 0, Sigma),
  nrow = N.Sim, ncol = N.Ahead)
Mat.Sim.2 = matrix(rep(rnorm(N.Sim, Mu, Sigma/sqrt(n.y-1)), N.Ahead),
  nrow = N.Sim, ncol = N.Ahead)
Mat.Sim = Mat.Sim.1 + Mat.Sim.2
Both = Kappa[n.y] + apply(Mat.Sim, 1, cumsum)
Sim.Q.Both = apply(Both, 1, Quantile)
# 
par(mar=c(4.2, 4, 1, 0.5), mgp=c(3, 1, 0), las=1, cex=1)
plot(Range, Sim.Q.Both[2, ], type = "l", lwd = 2, axes = FALSE,
     xlab = "Year", ylab = expression(kappa),
     ylim = range(Sim.Q.Both, Sim.Q.Both))
axis(1, las = 1, at = seq(2010, 2040, by = 5), tcl = -0.4)
axis(2, seq(-70, -10, by = 5), tcl = -0.4)
lines(Range, Sim.Q.Both[1, ], lwd = 2)
lines(Range, Sim.Q.Both[3, ], lwd = 2)
#
lines(Range, Central, lwd = 2, lty = 2)
VAR1 = (1:N.Ahead) * Sigma2
VAR2  = (1:N.Ahead)^2/(n.y-1) * Sigma2
SE.Both = sqrt(VAR1 + VAR2)
PI.CI.Up = Central + 1.96*SE.Both
PI.CI.Dn = Central - 1.96*SE.Both
lines(Range, PI.CI.Up, lty = 2, lwd = 2)
lines(Range, PI.CI.Dn, lty = 2, lwd = 2)
legend("bottomleft", legend = c("95%: theoretical", "95%: simulated"),
       lty = c(2,1), lwd = 2, bty = "n")
